#include "main.h"
#include "lcd_drv.h"


#define T60US	(u16)(XTAL / 5 * 60e-6 + 1)
#define T1US	(u16)(XTAL / 5 * 1e-6 + 1)


static u8 lcd_e01;			// controller select


static void lcd_wait( u8 time )		// tw = 1us + (n-1) * 60us
{
  u8 i = T1US;				// 1us

  do{
    do{
      asm volatile("nop");
      asm volatile("nop");		// 5 cycle loop to support 20MHz
    }while( --i );
    i = T60US;				// 60us
  }while( --time );
}


static void lcd_nibble( u8 d )
{
  LCD_D4 = 0; if( d & 1<<4 ) LCD_D4 = 1;
  LCD_D5 = 0; if( d & 1<<5 ) LCD_D5 = 1;
  LCD_D6 = 0; if( d & 1<<6 ) LCD_D6 = 1;
  LCD_D7 = 0; if( d & 1<<7 ) LCD_D7 = 1;

  if( lcd_e01 & 1<<0 ) LCD_E0 = 1;
  if( lcd_e01 & 1<<1 ) LCD_E1 = 1;
  lcd_wait( 1 );			// 1us
  LCD_E0 = 0;
  LCD_E1 = 0;
}


static void lcd_byte( u8 d )
{
  lcd_nibble( d );
  lcd_nibble( d<<4 );
  lcd_wait( 2 );			// 61us
}


void lcd_command( u8 d )
{
  LCD_RS = 0;
  lcd_byte( d );
  switch( d ){
    case 1:
    case 2:
    case 3: lcd_wait( 1600 / 60 + 2 );	// wait 1.6ms
  }
}


void lcd_data( u8 d )
{
  LCD_RS = 1;
  lcd_byte( d );
}


void lcd_text( u8 *t )
{
  while( *t ){
    lcd_data( *t );
    t++;
  }
}


void lcd_init( void )
{
  LCD_PORT &= ~LCD_LINES;
  LCD_DDR |= LCD_LINES;
  LCD_RS = 0;				// send commands for initialization
  lcd_wait( 15000 / 60 + 2 );		// wait 15ms

  lcd_e01 = 3;				// initialize both controller

  lcd_nibble( 0x30 );
  lcd_wait( 4100 / 60 + 2 );		// wait 4.1ms

  lcd_nibble( 0x30 );
  lcd_wait( 3 );			// wait >100s

  lcd_nibble( 0x30 );			// 8 bit mode
  lcd_wait( 3 );			// wait >100s

  lcd_nibble( 0x20 );			// 4 bit mode
  lcd_wait( 3 );			// wait >100s

  lcd_command( 0x28 );			// 2 lines 5*7
  lcd_command( 0x08 );			// display off
  lcd_command( 0x01 );			// display clear
  lcd_command( 0x06 );
  lcd_command( 0x0C );			// no cursor, no blink
  lcd_e01 = 1<<0;			// enable 1. controller
}


void lcd_pos( u8 line, u8 column )
{
  lcd_e01 = 1;				// 1./2. line
  if( line & 2 )
    lcd_e01 = 2;			// 3./4. line

  if( line & 1 )
    column += 64;

  lcd_command( 0x80 + column );
}
